// /!\ Autogenerated file - do not modify

#define F3D_PLUGIN_IS_STATIC 1

#include "plugin.h"
#include "plugin_export.h"

#include <memory>

#include "reader_FBX.h"
#include "reader_COLLADA.h"
#include "reader_DXF.h"
#include "reader_OFF.h"
#include "reader_DirectX.h"
#include "reader_3MF.h"


static std::shared_ptr<f3d::plugin> gPlugin;

#if F3D_PLUGIN_IS_STATIC
f3d::plugin* init_plugin_static_assimp()
#else
extern "C" F3D_PLUGIN_EXPORT f3d::plugin* init_plugin()
#endif
{
  if (!gPlugin)
  {
    gPlugin = std::make_shared<f3d::plugin>(
      "assimp",
      "Assimp support (version 6.0.5)",
      "1.0",
      std::vector<std::shared_ptr<f3d::reader>>{ std::make_shared<reader_FBX>(),std::make_shared<reader_COLLADA>(),std::make_shared<reader_DXF>(),std::make_shared<reader_OFF>(),std::make_shared<reader_DirectX>(),std::make_shared<reader_3MF>(), }
    );
  }

  return gPlugin.get();
}
